#define __KERNEL__
#define MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>

#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/version.h>
#include <linux/brlock.h>

#define ACK_QUEUE_LEN	8

#define tcpseq		h.th->seq
#define tcpackseq	h.th->ack_seq
#define tcpwin		h.th->window
#define ipid		nh.iph->id

static struct nf_hook_ops ack_hook;
static unsigned int ack_seq_ack[ACK_QUEUE_LEN];
static unsigned int make_unique_unsigned(struct sk_buff *);

static unsigned int ack_adaa(	unsigned int hooknum,
				struct sk_buff **skb,
				const struct net_device *in,
				const struct net_device *out,
				int (*okfn)(struct sk_buff *))
	{
	static unsigned int target;
	register unsigned int i, check;

	if((*skb)->nh.iph->protocol !=IPPROTO_TCP)
		return NF_ACCEPT;

	if((*skb)->h.th->ack && !(*skb)->h.th->syn && !(*skb)->h.th->rst)
		return NF_ACCEPT;

	check =make_unique_unsigned(*skb);

	for(i =0x0000; i !=ACK_QUEUE_LEN; i++)
		if(ack_seq_ack[i] ==check)
			return NF_DROP;

	ack_seq_ack[target] =check;

	if(target ==ACK_QUEUE_LEN -1)
		target =0x0000;
	else
		target++;

	return NF_ACCEPT;	
	}

int init_module(void)
	{
	int ret;

	ack_hook.list.next =NULL;
	ack_hook.list.prev =NULL;
	ack_hook.pf =PF_INET;
	ack_hook.hook =ack_adaa;
	ack_hook.hooknum =NF_IP_LOCAL_IN;

	printk(KERN_INFO "mod for fix dup ack attack - by vecna@s0ftpj.org\n");

	if(ret =nf_register_hook(&ack_hook))
		printk(KERN_INFO "unable to add ack_hook on input flow\n");

	return ret;
	}

void cleanup_module(void)
	{
	nf_unregister_hook(&ack_hook);

	printk(KERN_INFO "unloading anti dup ack attack module\n");
	}

static unsigned int make_unique_unsigned(struct sk_buff *skb)
	{
	unsigned int intermediate;

	memcpy(&intermediate, &skb->ipid, 2);
	memcpy(((char *)&intermediate +2), &skb->tcpwin, 2);

	return (skb->tcpseq ^ skb->tcpackseq ^ intermediate);
	}
